/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Functions managing maps
 *
 * $Id: map.c,v 1.121 2009/05/01 04:50:50 zlb Exp $ */

#define NEED_GetVariableArgs

#include "phg.h"
#include "periodic.h"

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

/* The new (much faster) algorithm using g->owner_xxxx_xxxx arrays.
 *
 * Note:
 *
 *   1. The unreferenced entries are not handled and map->D2Lmap is used to
 *	remove them from the map (with localsize adjusted accordingly)
 *
 *   2. When there are periodic boundaries, map->D2Lmap is also used to
 *	remove duplicate entries. */

static void
build_periodic_D2Lmap(MAP *map)
/* builds map->D2Lmap for periodic boundaries, 
 *
 * Each group of periodic vertices is designated a local owner (which
 * matches the global ownership of the vertex). If a vertex is not the
 * local owner of its group then its D2Lmap entry is negated (-1 -.).
 *
 * This function also sets up map->dofsize. */
{
    GRID *g = map->dofs[0]->g;
    DOF *u;
    DOF_TYPE *type;
    INT i, j, k, l, m, dindex, lindex;
    int ii, jj, count;

    if (g->period == NULL) {
	map->dofsize = map->localsize;
	return;
    }

#if DEBUG
    if (g->nvert > 0) {
	assert(g->owner_index_vert != NULL && g->owner_rank_vert != NULL);
	assert(g->period != NULL && g->period->ordering != NULL);
    }
#endif	/* DEBUG */

    map->D2Lmap = phgAlloc(map->localsize * sizeof(*map->D2Lmap));

    dindex = 0;
    lindex = 0;
    for (ii = 0; ii < map->ndof; ii++) {
	u = map->dofs[ii];
	if ((type = u->type) == NULL)
	    type = u->hp->info->types[u->hp->max_order];

	/* vertex DOF */
	if (type->np_vert > 0) {
	    count = u->count_vert;
	    i = 0;
	    while (i < g->nvert) {
		m = g->period->L2Gmap_vert[g->period->ordering[i]];
		l = i;
		/* choose between identical vertices the one having the
		 * OWNER flag, or the one with the smallest index */
		for (k = i; k < g->nvert; k++) {
		    if (g->period->L2Gmap_vert[g->period->ordering[k]] != m)
			break;
		    if ((g->types_vert[g->period->ordering[k]] & OWNER))
			l = k;
		}
		/* change local owner to a referenced one (if one exists) */
		while (l < k &&
			g->types_vert[g->period->ordering[l]] == UNREFERENCED)
		    l++;
		if (l >= k) {
		    /* all vertices in the group are unreferenced */
		    l = i;
		}
		
		/* set D2Lmap, the non-owner entries are negated. */
		for (; i < k; i++) {
		    j = dindex + count * g->period->ordering[i];
		    if (i == l) {
			for (jj = 0; jj < count; jj++)
			    map->D2Lmap[j + jj] = lindex + jj;
		    }
		    else {
			for (jj = 0; jj < count; jj++)
			    map->D2Lmap[j + jj] = -1 - lindex - jj;
		    }
		}
		lindex += count;
	    }
	    dindex += count * g->nvert;
	}

	/* edge DOF */
	if (type->np_edge > 0) {
	    count = u->count_edge;
	    for (i = 0; i < g->nedge; i++) {
		if (DofIsHP(u))
		    count = (u->hp->edge_index[i + 1] - u->hp->edge_index[i])
				* u->dim;
		for (jj = 0; jj < count; jj++)
		    map->D2Lmap[dindex++] = lindex++;
	    }
	}

	/* face DOF */
	if (type->np_face > 0) {
	    count = u->count_face;
	    for (i = 0; i < g->nface; i++) {
		if (DofIsHP(u))
		    count = (u->hp->face_index[i + 1] - u->hp->face_index[i])
				* u->dim;
		for (jj = 0; jj < count; jj++)
		    map->D2Lmap[dindex++] = lindex++;
	    }
	}

	/* elem DOF */
	if (type->np_elem > 0) {
	    count = u->count_elem;
	    for (i = 0; i < g->nelem; i++) {
		if (DofIsHP(u))
		    count = (u->hp->elem_index[i + 1] - u->hp->elem_index[i])
				* u->dim;
		for (jj = 0; jj < count; jj++)
		    map->D2Lmap[dindex++] = lindex++;
	    }
	}
    }

    assert(map->localsize == dindex);
    map->nlocal = map->localsize = lindex;
    map->dofsize = dindex;

    return;
}

static void
build_maps(MAP *map)
{
    int ii;
    DOF *u = map->dofs[0];
    GRID *g = u->g;
    INT i, j;
#if USE_MPI
    INT dindex, lindex, vindex;
    BOOLEAN vert_flag, edge_flag, face_flag, elem_flag;
    int rank = 0, nsend, nrecv, count, n, jj;
    int *scnts = NULL, *sdsps = NULL, *rcnts = NULL, *rdsps = NULL;
    INT oindex, *sbuf, *rbuf, *buf;
    INT *index_vert, *index_edge, *index_face, *index_elem;
    SHORT *rank_vert, *rank_edge, *rank_face, *rank_elem;
    /* list for recording information about non owned entries, used
     * to avoid multiple traversal of vertices/edges/faces/elements */
    struct {
	INT	*lindices;	/* indices in L2Vmap[] */
	INT	index;		/* vertex/edge/face/element index */
	GTYPE	gtype;		/* geometric type */
    } *nonowned = NULL, *no = NULL;
    INT *nonowned_buffer = NULL, nonowned_count = 0;
    BOOLEAN *flags_vert, *flags_edge, *flags_face, *flags_elem;
    BOOLEAN have_unreferenced;
    DOF_TYPE *type;
#endif	/* USE_MPI */

    map->O2Gmap = map->L2Vmap = map->D2Lmap = NULL;
    map->partition = phgAlloc((g->nprocs + 1) * sizeof(*map->partition));
    map->offset = phgAlloc(((int)map->ndof + 1) * sizeof(*map->offset));
    map->offset[0] = 0;
    for (ii = 0; ii < map->ndof; ii++) {
	map->offset[ii + 1] = map->offset[ii] + DofGetDataCount(map->dofs[ii]);
    }

    build_periodic_D2Lmap(map);

    if (g->nprocs == 1) {
	map->partition[0] = 0;
	map->partition[1] = map->nlocal = map->nglobal = map->localsize;
	map->lif = 1.0;
	if (map->D2Lmap != NULL && map->dofsize > map->localsize) {
	    /* restore negated indices in D2Lmap */
	    for (i = 0; i < map->dofsize; i++) {
		if ((j = map->D2Lmap[i]) < 0) {
		    map->D2Lmap[i] = -1 - j;
		    assert((j = map->D2Lmap[i]) >= 0 && j < map->localsize);
		}
	    }
	}
	return;
    }

#if USE_MPI
    assert(g->nprocs > 1);
    vert_flag = edge_flag = face_flag = elem_flag = FALSE;
    for (ii = 0; ii < map->ndof; ii++) {
	u = map->dofs[ii];
	if ((type = u->type) == NULL)
	    type = u->hp->info->types[u->hp->max_order];
	if (type->np_vert > 0)
	    vert_flag = TRUE;
	if (type->np_edge > 0)
	    edge_flag = TRUE;
	if (type->np_face > 0)
	    face_flag = TRUE;
	if (type->np_elem > 0)
	    elem_flag = TRUE;
    }

    index_vert = g->owner_index_vert;
    index_edge = g->owner_index_edge;
    index_face = g->owner_index_face;
    index_elem = g->owner_index_elem;

    rank_vert = g->owner_rank_vert;
    rank_edge = g->owner_rank_edge;
    rank_face = g->owner_rank_face;
    rank_elem = g->owner_rank_elem;

    /* create owner rank/index mapping if not available */
    if (face_flag) {
	if (index_face == NULL) {
	    assert(rank_face == NULL);
	    index_face = phgAlloc(g->nface * sizeof(*index_face));
	    rank_face = phgAlloc(g->nface * sizeof(*rank_face));
	    for (i = 0; i < g->nface; i++) {
		index_face[i] = -1;
		rank_face[i] = -1;
	    }
	    for (i = 0; i < g->neighbours.count; i++) {
		RNEIGHBOUR *rn = g->neighbours.list + i;
		lindex = rn->local->faces[rn->vertex];
		if ((g->types_face[lindex] & OWNER))
		    continue;
		index_face[lindex] = rn->peer_face;
		rank_face[lindex] = rn->rank;
	    }
	}
	else {
	    assert(rank_face != NULL);
	}
    }

    scnts = phgAlloc(4 * g->nprocs * sizeof(*scnts));
    sdsps = scnts + g->nprocs;
    rcnts = sdsps + g->nprocs;
    rdsps = rcnts + g->nprocs;
    memset(scnts, 0, 2 * g->nprocs * sizeof(*scnts));

    nonowned_count = (vert_flag ? g->nvert - g->nvert_owned : 0) +
		     (edge_flag ? g->nedge - g->nedge_owned : 0) +
		     (face_flag ? g->nface - g->nface_owned : 0) +
		     (elem_flag ? g->nelem - g->nelem_owned : 0);
    nonowned = phgAlloc(nonowned_count * sizeof(*nonowned));
    nonowned_buffer = phgAlloc(nonowned_count * map->ndof
					* sizeof(*nonowned_buffer));
    for (i = 0; i < nonowned_count; i++)
	nonowned[i].lindices = nonowned_buffer + i * map->ndof;

    map->L2Vmap = phgAlloc(map->localsize * sizeof(*map->L2Vmap));

    dindex = lindex = vindex = 0;
    oindex = map->localsize;

    flags_vert = (vert_flag ? phgCalloc(g->nvert, sizeof(*flags_vert)) : NULL);
    flags_edge = (edge_flag ? phgCalloc(g->nedge, sizeof(*flags_edge)) : NULL);
    flags_face = (face_flag ? phgCalloc(g->nface, sizeof(*flags_face)) : NULL);
    flags_elem = (elem_flag ? phgCalloc(g->nelem, sizeof(*flags_elem)) : NULL);

    for (ii = 0; ii < map->ndof; ii++) {
	u = map->dofs[ii];
	phgDebug((4, "processing DOF %s:\n", u->name));
	no = nonowned;

	/* vertex DOF */
	if (vert_flag) {
	    count = u->count_vert;
	    for (i = 0; i < g->nvert; i++, dindex += count) {
		if (map->D2Lmap != NULL && map->D2Lmap[dindex] < 0)
		    continue;
		if (g->types_vert[i] & OWNER) {
		    phgDebug((4, "vert %d, lindex=%d, vindex=%d\n",
							i, lindex, vindex));
		    if (g->period == NULL) {
			for (jj = 0; jj < count; jj++)
			    map->L2Vmap[lindex++] = vindex++;
		    }
		    else {
			j = map->D2Lmap[dindex];
			for (jj = 0; jj < count; jj++)
			    map->L2Vmap[j++] = vindex++;
			lindex += count;
		    }
		}
		else {
		    if (flags_vert[i] == 0) {
			no->index = i;
			no->gtype = VERTEX;
			if (rank_vert != NULL && (n = rank_vert[i]) >= 0)
			    scnts[2 * n]++;
			flags_vert[i] = 1;
		    }
		    if (g->period == NULL) {
			no->lindices[ii] = lindex;
			for (jj = 0; jj < count; jj++)
			    map->L2Vmap[lindex++] = oindex++;
		    }
		    else {
			j = map->D2Lmap[dindex];
			no->lindices[ii] = j;
			for (jj = 0; jj < count; jj++)
			    map->L2Vmap[j++] = oindex++;
			lindex += count;
		    }
		    no++;
		}
	    }
	}

	/* edge DOF */
	if (edge_flag) {
	    count = u->count_edge;
	    for (i = 0; i < g->nedge; i++) {
		if (DofIsHP(u))
		    count = (u->hp->edge_index[i + 1] - u->hp->edge_index[i])
				* u->dim;
		if (g->types_edge[i] & OWNER) {
		    phgDebug((4, "edge %d, lindex=%d, vindex=%d\n",
							i, lindex, vindex));
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = vindex++;
		}
		else {
		    if (flags_edge[i] == 0) {
			no->index = i;
			no->gtype = EDGE;
			if (rank_edge != NULL && (n = rank_edge[i]) >= 0)
			    scnts[2 * n]++;
			flags_edge[i] = 1;
		    }
		    no->lindices[ii] = lindex;
		    no++;
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = oindex++;
		}
	    }
	    dindex += DofGetEdgeDataCount(u);
	}

	/* face DOF */
	if (face_flag) {
	    count = u->count_face;
	    for (i = 0; i < g->nface; i++) {
		if (DofIsHP(u))
		    count = (u->hp->face_index[i + 1] - u->hp->face_index[i])
				* u->dim;
		if (g->types_face[i] & OWNER) {
		    phgDebug((4, "face %d, lindex=%d, vindex=%d\n",
							i, lindex, vindex));
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = vindex++;
		}
		else {
		    if (flags_face[i] == 0) {
			no->index = i;
			no->gtype = FACE;
			if (rank_face != NULL && (n = rank_face[i]) >= 0)
			    scnts[2 * n]++;
			flags_face[i] = 1;
		    }
		    no->lindices[ii] = lindex;
		    no++;
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = oindex++;
		}
	    }
	    dindex += DofGetFaceDataCount(u);
	}

	/* elem DOF */
	if (elem_flag) {
	    count = u->count_elem;
	    for (i = 0; i < g->nelem; i++) {
		if (DofIsHP(u))
		    count = (u->hp->elem_index[i + 1] - u->hp->elem_index[i])
				* u->dim;
		if (g->types_elem[i] != UNREFERENCED) {
		    phgDebug((4, "elem %d, lindex=%d, vindex=%d\n",
							i, lindex, vindex));
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = vindex++;
		}
		else {
		    if (flags_elem[i] == 0) {
			no->index = i;
			no->gtype = ELEMENT;
			if (rank_elem != NULL && (n = rank_elem[i]) >= 0)
			    scnts[2 * n]++;
			flags_elem[i] = 1;
		    }
		    no->lindices[ii] = lindex;
		    no++;
		    for (jj = 0; jj < count; jj++)
			map->L2Vmap[lindex++] = oindex++;
		}
	    }
	    dindex += DofGetElementDataCount(u);
	}
    }
    phgFree(flags_vert);
    phgFree(flags_edge);
    phgFree(flags_face);
    phgFree(flags_elem);
    map->nlocal = vindex;
    assert(map->dofsize == dindex);
    assert(map->localsize == lindex);
    assert(map->nlocal <= map->localsize);
    assert(no - nonowned <= nonowned_count);
    nonowned_count = no - nonowned;

    /* restore negated indices in D2Lmap */
    if (map->D2Lmap != NULL && map->dofsize > map->localsize) {
	for (i = 0; i < map->dofsize; i++) {
	    if ((j = map->D2Lmap[i]) < 0) {
		map->D2Lmap[i] = -1 - j;
		assert((j = map->D2Lmap[i]) >= 0 && j < map->localsize);
	    }
	}
    }

    map->O2Gmap = phgAlloc((map->localsize - map->nlocal)
					* sizeof(*map->O2Gmap));

    /* put nlocal in scnts[2k+1] */
    for (ii = 0; ii < g->nprocs; ii++)
	scnts[2 * ii + 1] = map->nlocal;
    MPI_Alltoall(scnts, 2, MPI_INT, rcnts, 2, MPI_INT, g->comm);

    /* now scnts[2k]/rcnts[2k] contain the send/recv counts,
     * scnts[2k+1]/rcnts[2k+1] contain nlocal's for all processes */
    map->partition[0] = 0;
    i = 0;
    for (ii = 0; ii < g->nprocs; ii++) {
	if (i < rcnts[2 * ii + 1])
	    i = rcnts[2 * ii + 1];
	map->partition[ii + 1] = map->partition[ii] + rcnts[2 * ii + 1];
	scnts[ii] = scnts[2 * ii];
	rcnts[ii] = rcnts[2 * ii];
    }
    if (g->period != NULL)
	map->nglobal = map->partition[g->nprocs];
    else
	assert(map->nglobal == map->partition[g->nprocs]);
    map->lif = (i * (FLOAT)g->nprocs) / map->nglobal;

    nsend = nrecv = 0;
    for (ii = 0; ii < g->nprocs; ii++) {
	sdsps[ii] = nsend;
	rdsps[ii] = nrecv;
	nsend += scnts[ii];
	nrecv += rcnts[ii];
    }

    /* allocate max(ndof,2) INTs for each entry in sbuf/rbuf, which contains:
     *     send: INT[0] = the index in the owner process, INT[1] = gtype
     *     recv: the global indices, one for each DOF */
    n = (map->ndof <= 1 ? 2 : map->ndof);
    sbuf = phgAlloc((nsend + nrecv) * n * sizeof(*sbuf));
    rbuf = sbuf + nsend * n;

    /* set up send buffer */
    for (i = 0, no = nonowned; i < nonowned_count; i++, no++) {
	switch (no->gtype) {
	    case VERTEX:
		if (rank_vert == NULL || (rank = rank_vert[no->index]) == -1)
		    break;
		ii = n * (sdsps[rank]++);
		sbuf[ii + 0] = index_vert[no->index];
		sbuf[ii + 1] = no->gtype;
		phgDebug((4, "send[%d]: vert %d ==> vert %d on proc %d\n",
			     sdsps[rank] - 1, no->index, sbuf[ii + 0], rank));
		break;
	    case EDGE:
		if (rank_edge == NULL || (rank = rank_edge[no->index]) == -1)
		    break;
		ii = n * (sdsps[rank]++);
		sbuf[ii + 0] = index_edge[no->index];
		sbuf[ii + 1] = no->gtype;
		phgDebug((4, "send[%d]: edge %d ==> edge %d on proc %d\n",
			     sdsps[rank] - 1, no->index, sbuf[ii + 0], rank));
		break;
	    case FACE:
		if (rank_face == NULL || (rank = rank_face[no->index]) == -1)
		    break;
		ii = n * (sdsps[rank]++);
		sbuf[ii + 0] = index_face[no->index];
		sbuf[ii + 1] = no->gtype;
		phgDebug((4, "send[%d]: face %d ==> face %d on proc %d\n",
			     sdsps[rank] - 1, no->index, sbuf[ii + 0], rank));
		break;
	    case ELEMENT:
		if (rank_elem == NULL || (rank = rank_elem[no->index]) == -1)
		    break;
		ii = n * (sdsps[rank]++);
		sbuf[ii + 0] = index_elem[no->index];
		sbuf[ii + 1] = no->gtype;
		phgDebug((4, "send[%d]: elem %d ==> elem %d on proc %d\n",
				no->index, sbuf[ii + 0], rank));
		break;
	    default:
		phgError(1, "%s:%d, unexpected.\n", __FILE__, __LINE__);
	}
    }
    phgDebug((4, "nsend = %d, nrecv = %d\n", nsend, nrecv));
    assert(sdsps[g->nprocs - 1] == nsend);

    /* restore sdsps */
    for (ii = 0; ii < g->nprocs; ii++)
	sdsps[ii] -= scnts[ii];

    {
	/* construct MPI Datatype. Since only the first two bytes need to be sent,
	 * datatype is defined as two consequent INTs, with an extent of n INTs */
	int blocklens[2];
	MPI_Aint indices[2];
	MPI_Datatype old_types[2], type;

	blocklens[0]	= 2;
	indices[0]	= 0;
	old_types[0]	= PHG_MPI_INT;

	blocklens[1]	= 1;
	indices[1]	= n * sizeof(INT);
	old_types[1]	= MPI_UB;

	MPI_Type_struct(2, blocklens, indices, old_types, &type);
	MPI_Type_commit(&type);
	phgAlltoallv(sbuf, scnts, sdsps, type, rbuf, rcnts, rdsps, type, g->comm);
	MPI_Type_free(&type);
    }

    /* process received lists */
    rank = 0;			/* source rank, for debugging */
    for (i = 0, buf = rbuf; i < nrecv; i++, buf += n) {
	GTYPE gtype = buf[1];
	assert(buf[1] >= VERTEX && buf[1] <= ELEMENT);
	lindex = buf[0];	/* local index of vertex/edge/face/element */
	while (rank < g->nprocs && i >= rdsps[rank] + rcnts[rank])
	    rank++;
	phgDebug((4, "Processing recv[%d] from p%d, gtype=%d, lindex=%d:\n",
			i, rank, gtype, lindex));
	assert(rank < g->nprocs);
	for (ii = 0; ii < map->ndof; ii++) {
	    u = map->dofs[ii];
	    phgDebug((4, "    DOF %s\n", u->name));
	    switch (gtype) {
		case VERTEX:
		    assert((g->types_vert[lindex] & OWNER));
		    count = u->count_vert;
		    vindex = map->offset[ii] + count * lindex;
		    break;
		case EDGE:
		    assert((g->types_edge[lindex] & OWNER));
		    vindex = map->offset[ii] + DofGetVertexDataCount(u);
		    if (!DofIsHP(u)) {
			count = u->count_edge;
			vindex += count * lindex;
		    }
		    else {
			count = (u->hp->edge_index[lindex + 1] -
				 u->hp->edge_index[lindex]) * u->dim;
			vindex += u->hp->edge_index[lindex] * u->dim;
		    }
		    break;
		case FACE:
		    assert((g->types_face[lindex] & OWNER));
		    vindex = map->offset[ii] + DofGetVertexDataCount(u)
					     + DofGetEdgeDataCount(u);
		    if (!DofIsHP(u)) {
			count = u->count_face;
			vindex += count * lindex;
		    }
		    else {
			count = (u->hp->face_index[lindex + 1] -
				 u->hp->face_index[lindex]) * u->dim;
			vindex += u->hp->face_index[lindex] * u->dim;
		    }
		    break;
		case ELEMENT:
		    assert((g->types_elem[lindex] & OWNER));
		    vindex = map->offset[ii] + DofGetVertexDataCount(u)
					     + DofGetEdgeDataCount(u)
					     + DofGetFaceDataCount(u);
		    if (!DofIsHP(u)) {
			count = u->count_elem;
			vindex += count * lindex;
		    }
		    else {
			count = (u->hp->elem_index[lindex + 1] -
				 u->hp->elem_index[lindex]) * u->dim;
			vindex += u->hp->elem_index[lindex] * u->dim;
		    }
		    break;
		default:
		    count = 0;		/* avoid gcc warning */
	    }
	    if (count == 0) {
		phgDebug((4, "\tno DOF data.\n"));
		buf[ii] = -1;
		continue;
	    }
	    if (map->D2Lmap != NULL) {
		assert(vindex >= 0 && vindex < map->dofsize);
		vindex = map->D2Lmap[vindex];
	    }
	    assert(vindex >= 0 && vindex < map->localsize);
	    phgDebug((4, "\tdof index=%d, vindex=%d, count=%d\n",
			  vindex, map->L2Vmap[vindex], count));
	    vindex = map->L2Vmap[vindex];
	    assert(vindex < map->nlocal);	/* must be owned by the proc */
	    buf[ii] = vindex + map->partition[g->rank];
	}
    }

    /* send back lists */
    {
	/* construct MPI Datatype. Since only the first two bytes need to be sent,
	 * datatype is defined as two consequent INTs, with an extent of n INTs */
	MPI_Datatype type;

	MPI_Type_contiguous(n, PHG_MPI_INT, &type);
	MPI_Type_commit(&type);
	phgAlltoallv(rbuf, rcnts, rdsps, type, sbuf, scnts, sdsps, type, g->comm);
	MPI_Type_free(&type);
    }

    /* process resulting buffer */
    have_unreferenced = FALSE;
    for (i = 0, no = nonowned; i < nonowned_count; i++, no++) {
	for (ii = 0; ii < map->ndof; ii++) {
	    u = map->dofs[ii];
	    switch (no->gtype) {
		case VERTEX:
		    count = u->count_vert;
		    rank = (rank_vert == NULL ? -1 : rank_vert[no->index]);
		    break;
		case EDGE:
		    if (!DofIsHP(u))
			count = u->count_edge;
		    else
			count = (u->hp->edge_index[no->index + 1] -
				 u->hp->edge_index[no->index]) * u->dim;
		    rank = (rank_edge == NULL ? -1 : rank_edge[no->index]);
		    break;
		case FACE:
		    if (!DofIsHP(u))
			count = u->count_face;
		    else
			count = (u->hp->face_index[no->index + 1] -
				 u->hp->face_index[no->index]) * u->dim;
		    rank = (rank_face == NULL ? -1 : rank_face[no->index]);
		    break;
		case ELEMENT:
		    if (!DofIsHP(u))
			count = u->count_elem;
		    else
			count = (u->hp->elem_index[no->index + 1] -
				 u->hp->elem_index[no->index]) * u->dim;
		    rank = (rank_elem == NULL ? -1 : rank_elem[no->index]);
		    break;
		default:
		    count = 0;		/* make gcc happy */
		    rank = -1;
		    phgError(-1, "%s:%d, unexpected.\n", __FILE__, __LINE__);
	    }
	    if (rank >= 0) {
		vindex = sbuf[n * sdsps[rank] + ii];
		assert(vindex >= 0 || count == 0);
	    }
	    else {
		have_unreferenced = TRUE;
	    }
	    lindex = no->lindices[ii];
	    for (jj = 0; jj < count; jj++) {
		map->L2Vmap[lindex] -= map->localsize - map->nlocal;
		oindex = map->L2Vmap[lindex++] - map->nlocal;
		if (rank >= 0)
		    map->O2Gmap[oindex] = vindex++;
		else
		    map->O2Gmap[oindex] = -1;
	    }
	}
	sdsps[rank]++;
    }

    phgFree(nonowned_buffer);
    phgFree(nonowned);
    phgFree(sbuf);
    phgFree(scnts);
    if (index_face != g->owner_index_face) {
	phgFree(index_face);
	phgFree(rank_face);
    }

    if (!have_unreferenced)
	return;

    /* adjust maps by removing unreferenced entries */
    nsend = map->localsize - map->nlocal;
    buf = phgAlloc(nsend * sizeof(*buf));
    for (i = 0, oindex = 0; i < nsend; i++) {
	if ((vindex = map->O2Gmap[i]) >= 0) {
	    buf[i] = oindex;
	    map->O2Gmap[oindex++] = vindex;
	}
	else {
	    buf[i] = -1;
	}
    }

    if (g->period == NULL) {
	/* setup map->L2Vmap and map->D2Lmap */
	assert(map->D2Lmap == NULL);
	map->D2Lmap = phgAlloc(map->dofsize * sizeof(*map->D2Lmap));
	for (i = 0, lindex = 0; i < map->localsize; i++) {
	    if ((vindex = map->L2Vmap[i]) < map->nlocal) {
		map->D2Lmap[i] = lindex;
		map->L2Vmap[lindex++] = vindex;
	    }
	    else {
		if ((vindex = buf[vindex - map->nlocal]) >= 0) {
		    map->D2Lmap[i] = lindex;
		    map->L2Vmap[lindex++] = vindex + map->nlocal;
		}
		else {
		    map->D2Lmap[i] = -1;
		}
	    }
	}
    }
    else {
	INT *L2Lmap;
	assert(map->D2Lmap != NULL);
	/* setup map->L2Vmap and save temperary local==>local map in L2Lmap */
	L2Lmap = phgAlloc(map->localsize * sizeof(*L2Lmap));
	for (i = 0, lindex = 0; i < map->localsize; i++) {
	    if ((vindex = map->L2Vmap[i]) < map->nlocal) {
		L2Lmap[i] = lindex;
		map->L2Vmap[lindex++] = vindex;
	    }
	    else {
		if ((vindex = buf[vindex - map->nlocal]) >= 0) {
		    L2Lmap[i] = lindex;
		    map->L2Vmap[lindex++] = vindex + map->nlocal;
		}
		else {
		    L2Lmap[i] = -1;
		}
	    }
	}

	/* update map->D2Lmap */
	for (i = 0; i < map->dofsize; i++)
	    map->D2Lmap[i] = L2Lmap[map->D2Lmap[i]];

	phgFree(L2Lmap);
    }

    /* update map->localsize */
    assert(lindex == map->nlocal + oindex);
    map->localsize = lindex;

    phgFree(buf);
#endif	/* USE_MPI */
}

INT
phgMapL2G(MAP *map, INT index)
{
    DOF *u = map->dofs[0];
    INT i;

    assert(index >= 0 && index < map->localsize);

    if (u->g->nprocs == 1)
	return index;

    i = (map->L2Vmap == NULL ? index : map->L2Vmap[index]);
    assert(i >= 0 && i < map->localsize);

    if (i < map->nlocal)
	return map->partition[map->rank] + i;
    else
	return map->O2Gmap[i - map->nlocal];
}

static void
get_bdry_map(MAP *map)
/* initialize map->bdry_* members */
{
    GRID *g;
    DOF *dof;
    int ii, n;
    INT *counts, count;
    INT i, j, k, index, index0;

    BTYPE *tmp_types[4];
    int tmp_count[4];
    INT tmp_size[4], *tmp_index[4];
    BTYPE DB_mask;

    if (map->dofs == NULL)
	return;

    g = map->dofs[0]->g;

    for (i = 0; i < map->ndof; i++)
        if (map->dofs[i]->DB_mask != 0)
	    break;
    if (i >= map->ndof)
	return;

    tmp_types[0] = g->types_vert;
    tmp_types[1] = g->types_edge;
    tmp_types[2] = g->types_face;
    tmp_types[3] = g->types_elem;

    tmp_size[0] = g->nvert;
    tmp_size[1] = g->nedge;
    tmp_size[2] = g->nface;
    tmp_size[3] = g->nelem;

    /* count number of boundary DOF owned by the present process.
     * Note: by putting the loop on map->dofs as the inner most one,
     * then entries in map->bdry[] are increasing */
    count = 0;
    phgFree(map->bdry);
    map->bdry = NULL;
    map->bdry_nlocal = 0;
    for (i = 0; i < map->ndof; i++) {
	dof = map->dofs[i];
	DB_mask = dof->DB_mask;
	if (DB_mask == 0)
	    continue;
	tmp_count[0] = dof->count_vert;
	tmp_index[0] = NULL;
	if (!DofIsHP(dof)) {
	    tmp_count[1] = dof->count_edge;
	    tmp_count[2] = dof->count_face;
	    tmp_count[3] = dof->count_elem;
	    tmp_index[1] = tmp_index[2] = tmp_index[3] = NULL;
	}
	else {
	    DOF_TYPE *type = dof->hp->info->types[dof->hp->max_order];
	    n = dof->dim;
	    tmp_count[1] = type->np_edge * n;
	    tmp_count[2] = type->np_face * n;
	    tmp_count[3] = type->np_elem * n;
	    tmp_index[1] = dof->hp->edge_index;
	    tmp_index[2] = dof->hp->face_index;
	    tmp_index[3] = dof->hp->elem_index;
	}

	index0 = 0;
	for (ii = 0; ii < 4; ii++) {
	    if ((n = tmp_count[ii]) == 0)
		continue;
	    for (j = 0; j < tmp_size[ii]; j++) {
		if (DofIsHP(dof) && tmp_index[ii] != NULL)
		    n = tmp_index[ii][j + 1] - tmp_index[ii][j];
		if (!(tmp_types[ii][j] & DB_mask)) {
		    index0 += n;
		    continue;
		}
		if (count + n > map->bdry_nlocal) {
		    map->bdry_nlocal = (count + n + 1024);
		    map->bdry = phgRealloc(map->bdry, map->bdry_nlocal *
							sizeof(*map->bdry));
		}
		for (index = index0; index < index0 + n; index++) {
		    k = phgMapD2L(map, i, index);
		    assert(k >= 0);
		    k = phgMapL2V(map, k);
		    assert(k < map->localsize);
#if 0
		    /* only mark local bdry entries */
		    if (k >= map->nlocal)
			continue;
#endif
		    map->bdry[count++] = k;
		}
		index0 += n;
	    }
	}
    }

    if (count > 0) {
	/* sort entries in bdry[] */
	qsort(map->bdry, count, sizeof(*map->bdry), phgCompINT);
	if (g->period != NULL) {
	    /* delete duplicate entries in map->bdry[] */
	    for (i = 0, k = 0; i < count;) {
		for (j = i + 1; j < count && map->bdry[j] == map->bdry[i]; j++);
		map->bdry[k++] = map->bdry[i];
		i = j;
	    }
	    count = k;
	}
    }

    map->bdry_localsize = count;
    for (i = 0; i < count && map->bdry[i] < map->nlocal; i++);
    map->bdry_nlocal = count = i;

    /* gather counts from all processes */
    counts = phgAlloc(g->nprocs * sizeof(*counts));
#if USE_MPI
    MPI_Allgather(&count, 1, PHG_MPI_INT, counts, 1, PHG_MPI_INT, g->comm);
    k = 0;
    for (i = 0; i < g->nprocs; i++) {
	map->bdry_partition[i] = k;
	k += counts[i];
    }
    map->bdry_partition[i] = k;
#else /* USE_MPI */
    k = counts[0] = count;
    map->bdry_partition[0] = 0;
    map->bdry_partition[1] = count;
#endif /* USE_MPI */
    map->bdry_nglobal = k;

    phgInfo(2, "nlocal: %d, boundary DOF: %d (%d total)\n",
	    map->nlocal, count, k);

    phgFree(counts);
}

MAP *
phgMapCreateSimpleMap(MPI_Comm comm, INT n, INT N)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    int i;

    assert(n != PHG_DECIDE || N != PHG_DECIDE);

    map->comm = comm;
#if USE_MPI
    MPI_Comm_size(comm, &map->nprocs);
    MPI_Comm_rank(comm, &map->rank);
#else
    map->nprocs = 1;
    map->rank = 0;
#endif	/* USE_MPI */

    if (n == PHG_DECIDE) {
	n = N / map->nprocs;
	if (map->rank < N % map->nprocs)
	    n++;
    }
    assert(N == PHG_DECIDE || n <= N);
    map->nlocal = n;
    map->partition = phgAlloc((map->nprocs + 1) * sizeof(*map->partition));
    map->partition[0] = 0;
#if USE_MPI
    MPI_Allgather(&n, 1, PHG_MPI_INT, map->partition + 1, 1, PHG_MPI_INT,
		  map->comm);
#else	/* USE_MPI */
    map->partition[1] = n;
#endif	/* USE_MPI */
    for (i = 0; i < map->nprocs; i++)
	map->partition[i+ 1] += map->partition[i];
    map->nglobal = map->partition[i];
    if (N == PHG_DECIDE)
	N = map->nglobal;

    if (map->nglobal > N) {
	/* special case: identical serial map on all processes */
	assert(map->nglobal == map->nprocs * N);
	map->nglobal = N;
#if USE_MPI
	map->comm = MPI_COMM_SELF;
#endif	/* USE_MPI */
	map->nprocs = 1;
	map->rank = 0;
	phgFree(map->partition);
	map->partition = phgAlloc(2 * sizeof(*map->partition));
	map->partition[0] = 0;
	map->partition[1] = N;
    }

    assert(map->nglobal == N);
    map->localsize = map->nlocal;

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(map->nprocs + 1, sizeof(*map->bdry_partition));

    return map;
}

#if USE_MPI
static INT *O2Gmap;

static int
O2Gmap_comp(const void *p0, const void *p1)
{
    return O2Gmap[*(const INT *)p0] - O2Gmap[*(const INT *)p1];
}

void
_phg_build_ordering(size_t size, INT map[], INT **ordering)
/* build array ordering[] such that map[ordering[i]] is increasing w.r.t. i */
{
    INT i;

    if (size <= 0) {
	*ordering = NULL;
	return;
    }
    if (*ordering != NULL)
	return;

    O2Gmap = map;
    *ordering = phgAlloc(size * sizeof(**ordering));
    for (i = 0; i < size; i++)
	(*ordering)[i] = i;
    qsort(*ordering, size, sizeof(**ordering), O2Gmap_comp);
}
#endif /* USE_MPI */

MAP *
phgMapCreateN(int ndof, DOF **dofs)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    GRID *g;
    DOF *u;
    int j;

    assert(dofs != NULL);

    map->ndof = ndof;
    map->dofs = phgAlloc(ndof * sizeof(*map->dofs));
    memcpy(map->dofs, dofs, ndof * sizeof(*map->dofs));
    u = dofs[0];
    g = u->g;

#if USE_MPI
    map->comm = g->comm;
#endif	/* USE_MPI */
    map->nprocs = g->nprocs;
    map->rank = g->rank;

    map->nlocal = 0;				/* updated in build_maps() */
    map->nglobal = map->localsize = 0;
    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];
	map->nglobal += DofGetDataCountGlobal(u);
	map->localsize += DofGetDataCount(u);
    }

    build_maps(map);

#if USE_MPI
    _phg_build_ordering(map->localsize - map->nlocal, map->O2Gmap,
							&map->ordering);
#endif	/* USE_MPI */

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(g->nprocs + 1, sizeof(*map->bdry_partition));
    get_bdry_map(map);

    return map;
}

MAP *
phgMapCreate(DOF *u, ...)
{
    int ndof;
    DOF **dofs;
    MAP *map;

    GetVariableArgs(ndof, dofs, u, DOF *);
    map = phgMapCreateN(ndof, dofs);
    phgFree(dofs);

    return map;
}

void
phgMapDestroy(MAP **map_ptr)
{
    MAP *map = *map_ptr;

    if (map == NULL)
	return;

    *map_ptr = NULL;

    phgMapDestroy(&map->map_nb);

    assert(map->refcount >= 0);

    if (map->refcount > 0) {
	map->refcount--;
	return;
    }

    phgFree(map->D2Lmap);
    phgFree(map->L2Vmap);
    phgFree(map->O2Gmap);
    phgFree(map->ordering);
    phgFree(map->partition);
    phgFree(map->offset);
    phgFree(map->bdry);
    phgFree(map->bdry_partition);

    phgFree(map->dofs);

#if USE_MPI
    phgMapDestroyCommInfo(&map->cinfo);
#endif	/* USE_MPI */

    phgFree(map);

    return;
}

MAP *
phgMapRemoveBoundaryEntries(MAP *map)
/* derives a new map from 'map' with all boundary entries removed */
{
    INT n, N;

    if (map->bdry_nglobal == 0) {
	map->refcount++;
	return map;
    }

    if (map->map_nb != NULL) {
	map->map_nb->refcount++;
	return map->map_nb;
    }
 
    n = map->nlocal - map->bdry_nlocal;
    N = map->nglobal - map->bdry_nglobal;
    assert(map->dofs != NULL);
    map->map_nb = phgMapCreateSimpleMap(map->comm, n, N);
    map->map_nb->refcount++;	/* note: map_nb is referenced by map */

    return map->map_nb;
}

/*--------- functions for vector scattering/gathering ---------*/

#if USE_MPI
COMM_INFO *
phgMapCreateCommInfo(MPI_Comm comm, int nprocs, int rank, INT nlocal,
		     INT localsize, INT *O2Gmap, INT *ordering, INT *partition)
/* Note: ordering may be NULL which means the natural ordering */
{
    COMM_INFO *cinfo;
    INT i, k, n, *work, O2Gsize = localsize - nlocal;
    int ii, mask;

    if (nprocs == 1)
	return NULL;

    cinfo = phgCalloc(1, sizeof(*cinfo));
    cinfo->comm = comm;

    cinfo->scnts = phgAlloc(4 * nprocs * sizeof(*cinfo->scnts));
    cinfo->sdsps = cinfo->scnts + nprocs;
    cinfo->rcnts = cinfo->sdsps + nprocs;
    cinfo->rdsps = cinfo->rcnts + nprocs;

    ii = 0;
    bzero(cinfo->rcnts, nprocs * sizeof(*cinfo->rcnts));
    mask = 0;
    if (ordering != NULL) {
	work = phgAlloc(O2Gsize * sizeof(*work));
	for (i = 0; i < O2Gsize; i++) {
	    k = work[i] = O2Gmap[ordering[i]];
	    while (ii < nprocs && partition[ii + 1] <= k)
		ii++;
	    assert(ii < nprocs && ii != rank);
	    cinfo->rcnts[ii]++;
	    mask = 1;
	}
    }
    else {
	work = O2Gmap;
	for (i = 0; i < O2Gsize; i++) {
	    k = O2Gmap[i];
	    while (ii < nprocs && partition[ii + 1] <= k)
		ii++;
	    assert(ii < nprocs && ii != rank);
	    cinfo->rcnts[ii]++;
	    mask = 1;
	}
    }

    /* attach mask to rcnts[] */
    if (mask != 0) {
	mask = (mask << (sizeof(mask) * 8 - 1));
	for (ii = 0; ii < nprocs; ii++)
	    cinfo->rcnts[ii] |= mask;
    }

    MPI_Alltoall(cinfo->rcnts, 1, MPI_INT, cinfo->scnts, 1, MPI_INT, comm);

    /* check and remove mask */
    mask = (1 << (sizeof(mask) * 8 - 1));
    n = 0;
    for (ii = 0; ii < nprocs; ii++) {
	if ((cinfo->scnts[ii] & mask))
	    n = 1;
	cinfo->scnts[ii] &= ~mask;
	cinfo->rcnts[ii] &= ~mask;
    }
    if (n == 0) {
	/* no communication needed for this map */
	assert(work == NULL);
	phgFree(cinfo->scnts);
	cinfo->scnts = cinfo->sdsps = cinfo->rcnts = cinfo->rdsps = NULL;
	cinfo->comm = MPI_COMM_NULL;
	return cinfo;
    }

    cinfo->size = nlocal;
    cinfo->rind = ordering;

    cinfo->rsize = cinfo->ssize = 0;
    for (ii = 0; ii < nprocs; ii++) {
	cinfo->rdsps[ii] = cinfo->rsize;
	cinfo->sdsps[ii] = cinfo->ssize;
	cinfo->rsize += cinfo->rcnts[ii];
	cinfo->ssize += cinfo->scnts[ii];
    }
    assert(cinfo->rsize == O2Gsize);
    /* prepare indices of data to send */
    cinfo->sind = phgAlloc(cinfo->ssize * sizeof(*cinfo->sind));
    phgAlltoallv(work,	      cinfo->rcnts, cinfo->rdsps, PHG_MPI_INT,
		 cinfo->sind, cinfo->scnts, cinfo->sdsps, PHG_MPI_INT, comm);
    if (work != O2Gmap)
	phgFree(work);
    /* convert indices in sind[] to local */
    n = partition[rank];
    for (i = 0; i < cinfo->ssize; i++) {
	assert(n <= cinfo->sind[i] && cinfo->sind[i] < partition[rank + 1]);
	cinfo->sind[i] -= n;
    }

    assert(cinfo->rcnts[rank] == 0 && cinfo->scnts[rank] == 0);

    return cinfo;
}

void
phgMapDestroyCommInfo(COMM_INFO **cinfo)
{
    if (*cinfo == NULL)
	return;

    assert((*cinfo)->nsends == 0 && (*cinfo)->nrecvs == 0
		&& (*cinfo)->buffer == NULL);

    phgFree((*cinfo)->scnts);
    phgFree((*cinfo)->sind);
    phgFree(*cinfo);

    *cinfo = NULL;
}

void
phgMapGatherBegin(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i;
    const INT *rind;
    FLOAT *sbuff, *rbuff, *p, *q;
    MPI_Datatype type;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    assert(cinfo->nsends == 0 && cinfo->nrecvs == 0 && cinfo->buffer == NULL);

    if (cinfo->ssize == 0 && cinfo->rsize == 0)
	return;

    if (cinfo->rind != NULL || nvec > 1) {
	/* Note: need rbuff to reorder the data if nvec > 1
	 * (FIXME: use MPI_Datatype with ex=1, size=rsize?) */
	cinfo->buffer = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize)
				      * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = sbuff + nvec * cinfo->ssize;
    }
    else {
	cinfo->buffer = phgAlloc(nvec * cinfo->ssize * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = offp_data;
    }

    if (nvec > 1) {
	MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	MPI_Type_commit(&type);
    }
    else {
	type = PHG_MPI_FLOAT;
    }

    /* copy data to send from offp_data to rbuff */
    if (cinfo->rind != NULL) {
	q = rbuff;
	rind = cinfo->rind;
	if (nvec == 1) {
	    /* faster code for nvec == 1 */
	    for (i = 0; i < cinfo->rsize; i++)
		*(q++) = offp_data[*(rind++)];
	}
	else {
	    for (i = 0; i < cinfo->rsize; i++) {
		p = offp_data + *(rind++);
		for (ii = 0; ii < nvec; ii++) {
		    *(q++) = *p;
		    p += cinfo->rsize;
		}
	    }
	}
    }
    else if (nvec > 1) {	/* no need to copy data if nvec == 1 */
	q = rbuff;
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data++;
	    for (ii = 0; ii < nvec; ii++) {
		*(q++) = *p;
		p += cinfo->rsize;
	    }
	}
    }

    phgAlltoallvBegin(rbuff, cinfo->rcnts, cinfo->rdsps, type,
		       sbuff, cinfo->scnts, cinfo->sdsps, type, cinfo->comm,
		       &cinfo->requests, &cinfo->nsends, &cinfo->nrecvs);

    if (nvec > 1)
	MPI_Type_free(&type);
}

void
phgMapGatherEnd(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i, *sind;
    FLOAT *p, *q;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    if (cinfo->nsends == 0 && cinfo->nrecvs == 0)
	return;

    phgAlltoallvEnd(cinfo->nsends, cinfo->nrecvs, &cinfo->requests);
    cinfo->nsends = cinfo->nrecvs = 0;

    p = cinfo->buffer;
    sind = cinfo->sind;
    if (nvec == 1) {
	/* faster code for nvec == 1 */
	for (i = 0; i < cinfo->ssize; i++) {
	    data[*(sind++)] += *(p++);
	}
    }
    else {
	for (i = 0; i < cinfo->ssize; i++) {
	    q = data + *(sind++);
	    for (ii = 0; ii < nvec; ii++) {
		*q += *(p++);
		q += cinfo->size;
	    }
	}
    }

    phgFree(cinfo->buffer);
    cinfo->buffer = NULL;
}

void
phgMapGather(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
/* data += offp_data */
{
    phgMapGatherBegin(cinfo, nvec, data, offp_data);
    phgMapGatherEnd(cinfo, nvec, data, offp_data);
}

void
phgMapScatterBegin(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i, *sind;
    FLOAT *sbuff, *rbuff, *p, *q;
    MPI_Datatype type;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    assert(cinfo->nsends == 0 && cinfo->nrecvs == 0 && cinfo->buffer == NULL);

    if (cinfo->ssize == 0 && cinfo->rsize == 0)
	return;

    if (nvec > 1) {
	static BOOLEAN warned = FALSE;
	if (!warned) {
	    phgWarning("function %s has not been tested!\n", __func__);
	    warned = TRUE;
	}
    }

    if (cinfo->rind != NULL || nvec > 1) {
	/* Note: need rbuff to reorder the data if nvec > 1
	 * (FIXME: use MPI_Datatype with ex=1, size=rsize?) */
	cinfo->buffer = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize)
				      * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = sbuff + nvec * cinfo->ssize;
    }
    else {
	cinfo->buffer = phgAlloc(nvec * cinfo->ssize * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = offp_data;
    }

    if (nvec > 1) {
	MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	MPI_Type_commit(&type);
    }
    else {
	type = PHG_MPI_FLOAT;
    }

    /* copy data to send from data to sbuff */
    p = sbuff;
    sind = cinfo->sind;
    if (nvec == 1) {
	/* faster code for nvec == 1 */
	for (i = 0; i < cinfo->ssize; i++) {
	    *(p++) = data[*(sind++)];
	}
    }
    else {
	for (i = 0; i < cinfo->ssize; i++) {
	    q = data + *(sind++);
	    for (ii = 0; ii < nvec; ii++) {
		*(p++) = *q;
		q += cinfo->size;
	    }
	}
    }

    phgAlltoallvBegin(sbuff, cinfo->scnts, cinfo->sdsps, type,
		       rbuff, cinfo->rcnts, cinfo->rdsps, type, cinfo->comm,
		       &cinfo->requests, &cinfo->nsends, &cinfo->nrecvs);

    if (nvec > 1)
	MPI_Type_free(&type);
}

void
phgMapScatterEnd(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i;
    const INT *rind;
    FLOAT *p, *q;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    if (cinfo->nsends == 0 && cinfo->nrecvs == 0)
	return;

    phgAlltoallvEnd(cinfo->nsends, cinfo->nrecvs, &cinfo->requests);
    cinfo->nsends = cinfo->nrecvs = 0;

    if (cinfo->rind != NULL) {
	q = cinfo->buffer + nvec * cinfo->ssize;
	rind = cinfo->rind;
	if (nvec == 1) {
	    /* faster code for nvec == 1 */
	    for (i = 0; i < cinfo->rsize; i++) {
		offp_data[*(rind++)] = *(q++);
	    }
	}
	else {
	    for (i = 0; i < cinfo->rsize; i++) {
		p = offp_data + *(rind++);
		for (ii = 0; ii < nvec; ii++) {
		    *p = *(q++);
		    p += cinfo->rsize;
		}
	    }
	}
    }
    else if (nvec > 1) {	/* data already in offp_data if nvec == 1 */
	q = cinfo->buffer + nvec * cinfo->ssize;
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data++;
	    for (ii = 0; ii < nvec; ii++) {
		*p = *(q++);
		p += cinfo->rsize;
	    }
	}
    }

    phgFree(cinfo->buffer);
    cinfo->buffer = NULL;
}

void
phgMapScatter(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
/* offp_data := data */
{
    phgMapScatterBegin(cinfo, nvec, data, offp_data);
    phgMapScatterEnd(cinfo, nvec, data, offp_data);
}
#endif	/* USE_MPI */

/*--------- functions for transferring data between DOF list and VEC --------*/

int
phgMapLocalDataToDof(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the solution back to DOFs.
 * 'vec' is the pointer to the local vector data (size = map->nlocal)
 */
{
#if 1	/*==================================================================*/
  /* The old circular algorithm */
  {
    DOF *u;
    INT i, j, k = 0, l, m, nmax;
    INT nlocal = map->nlocal;
    /* struct for storing remote data */
    struct {
	FLOAT value;
	INT gindex;
    } *rdata = NULL;
    INT *offp_counts = NULL;
#if USE_MPI
    MPI_Status status;
    MPI_Datatype type;
    INT first, last;
    int src, dst;

    if (map->nprocs > 1) {
	offp_counts = phgAlloc(map->nprocs * sizeof(*offp_counts));
	m = map->localsize - nlocal;
	MPI_Allgather(&m, 1, PHG_MPI_INT, offp_counts, 1, PHG_MPI_INT,
		      map->comm);
    }
#endif

    nmax = 0;
    if (offp_counts != NULL) {
	for (i = 0; i < map->nprocs; i++) {
	    if (nmax < offp_counts[i])
		nmax = offp_counts[i];
	}
    }
    if (nmax == 0) {
	k = 0;
	goto skip_comm;
    }

#if USE_MPI
    /* TODO: better use Alltoallv */

    /* copy indices of missing data */
    rdata = phgAlloc(2 * nmax * sizeof(*rdata));
    m = map->localsize - map->nlocal;
    for (i = 0; i < m; i++)
	rdata[i].gindex = map->O2Gmap[i];

    /* circulate lists of missing data and collecting missing data */
    j = map->rank;
    k = 0;
    dst = (map->rank == 0) ? map->nprocs - 1 : map->rank - 1;
    src = (map->rank == map->nprocs - 1) ? 0 : map->rank + 1;
    MPI_Type_contiguous(sizeof(*rdata), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    first = map->partition[map->rank];
    last = first + nlocal;
    for (i = 0; i < map->nprocs; i++) {
	if ((l = j + 1) >= map->nprocs)
	    l = 0;
	MPI_Sendrecv(rdata + k, offp_counts[j], type, dst, 111,
		     rdata + (nmax - k), offp_counts[l], type, src,
		     111, map->comm, &status);
	j = l;
	k = nmax - k;
	if (i == map->nprocs - 1)
	    break;
	/* pick up data from local data */
	for (l = 0; l < offp_counts[j]; l++) {
	    if ((m = rdata[k + l].gindex) < first)
		continue;
	    if (m >= last)
		continue;
	    rdata[k + l].value = data[m - first];
	}
    }
    MPI_Type_free(&type);
#endif /* USE_MPI */

  skip_comm:
    /* copy data to DOFs */
    for (l = 0; l < ndof; l++) {
	u = dofs[l];
	m = DofGetDataCount(u);
	phgDofClearCache(NULL, u, NULL, NULL, FALSE);
	for (i = 0; i < m; i++) {
	    j = phgMapD2L(map, l, i);
	    if (j < 0)
		u->data[i] = 0.0;
	    else if ((j = phgMapL2V(map, j)) < nlocal)
		u->data[i] = data[j];
	    else
		u->data[i] = rdata[k + j - nlocal].value;
	}
    }

    phgFree(offp_counts);
    phgFree(rdata);

    return 0;
  }
#endif	/*==================================================================*/
	
#if 0
  {
    int id;
    VEC *vec = phgMapCreateVec(map, 0);
    DOF ***pdofs = phgAlloc(ndof * sizeof(*pdofs));

    for (id = 0; id < ndof; id++)
	pdofs[id] = &dofs[id];
    vec->nvec = 1;
    vec->data = data;
    phgMapVecToDofArraysN(map, vec, FALSE, ndof, pdofs);
    vec->nvec = 0;
    vec->data = NULL;
    phgVecDestroy(&vec);
    phgFree(pdofs);
  }
#else	/* 0|1 */
  {
    int id;
    INT i, k, m;
    FLOAT *offp_data;
    DOF *u;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);

    offp_data = phgAlloc((map->localsize - map->nlocal) * sizeof(*data));

#if USE_MPI
    /* fetch remote data */
    if (map->nprocs > 1) {
	if (map->cinfo == NULL)
	    map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);
	phgMapScatter(map->cinfo, 1, data, offp_data);
    }
#endif	/* USE_MPI */

    for (id = 0; id < ndof; id++, dofs++) {
	u = dofs[id];
	m = DofGetDataCount(u);
	phgDofClearCache(NULL, u, NULL, NULL, FALSE);
	for (i = 0; i < m; i++) {
	    k = phgMapD2L(map, id, i);
	    if (k < 0)
		u->data[i] = 0.0;
	    else if ((k = phgMapL2V(map, k)) < map->nlocal)
		u->data[i] = data[k];
	    else
		u->data[i] = offp_data[k - map->nlocal];
	}
    }

    phgFree(offp_data);
  }
#endif	/* 0|1 */

    return 0;
}

int
phgMapDofToLocalData(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the data in the DOFs to local vector 
 * array (for e.g., setting up an initial solution of a solver).
 * 'data' is the pointer to the local data (size = map->nlocal)
 */
{
    DOF *u;
    int k;
    INT i, j, m;
    INT nlocal = map->nlocal;

    for (k = 0; k < ndof; k++) {
	u = dofs[k];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    if ((j = phgMapD2L(map, k, i)) >= 0 &&
		(j = phgMapL2V(map, j)) < nlocal)
		data[j] = u->data[i];
	}
    }

    return 0;
}

VEC *
phgMapDofArraysToVecN(MAP *map, int nvec, BOOLEAN remove_bdry,
			 VEC **vecptr, int ndof, DOF ***dofs)
/* copies data from an array of DOFs to a vector */
{
    VEC *vec = (vecptr == NULL ? NULL : *vecptr);
    int id, iv;
    DOF *u, *u0;
    INT i, k, m, *table;
    FLOAT *data;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    MagicCheck(VEC, vec);

    if (vec != NULL) {
	phgVecDestroy(&vec);
	if (vecptr != NULL)
	    *vecptr = NULL;
    }

    if (nvec <= 0)
	return vec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry) {
	/* no boundary entries */
	vec = phgMapCreateVec(map, nvec);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	MAP *map0 = phgMapRemoveBoundaryEntries(map);
	vec = phgMapCreateVec(map0, nvec);
	phgMapDestroy(&map0);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

    if (vecptr != NULL)
	*vecptr = vec;

    for (id = 0; id < ndof; id++, dofs++) {
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    if ((k = phgMapD2L(map, id, i)) < 0)
		continue;
	    k = table[phgMapL2V(map, k)];
	    if (k == -1 || k >= vec->map->nlocal)
		continue;
	    data = vec->data + k;
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		*data = u->data[i];
		data += vec->map->nlocal;
	    }
	}
    }
    phgFree(table);

    return vec;
}

VEC *
phgMapDofArraysToVec(MAP *map, int nvec, BOOLEAN remove_bdry,
			VEC **vecptr, DOF **u, ...)
/* copies data from an array of DOFs to a vector */
{
    int ndof;
    VEC *vec;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    vec = phgMapDofArraysToVecN(map, nvec, remove_bdry, vecptr, ndof, dofs);
    phgFree(dofs);
    return vec;
}

void
phgMapVecToDofArraysN(MAP *map, VEC *vec, BOOLEAN remove_bdry,
			 int ndof, DOF ***dofs)
/* copies data from a vector to an array of DOFs */
{
    int id, iv, nvec;
    INT i, j, k, m, stride;
    FLOAT *data, *offp_data = NULL;
    static FLOAT zero = 0.0;
    DOF *u0, *u;
    INT *table;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    assert(vec != NULL);
    MagicCheck(VEC, vec);

    if (vec->nvec <= 0)
	return;

    if (!vec->assembled)
	phgVecAssemble(vec);

    nvec = vec->nvec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry || vec->map->nglobal == map->nglobal) {
	/* no boundary entries */
	assert(vec->map->nglobal   == map->nglobal &&
	       vec->map->localsize == map->localsize);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	assert(vec->map->nlocal    == map->nlocal    - map->bdry_nlocal &&
	       vec->map->nglobal   == map->nglobal   - map->bdry_nglobal);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

#if USE_MPI
    /* fetch remote data */
    if (map->nprocs > 1 && map->cinfo == NULL)
	map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);

    if (map->cinfo != NULL && map->cinfo->comm != MPI_COMM_NULL) {
	/* collect off-process data */
	int ii;
	FLOAT *sbuff, *rbuff, *p, *q;
	MPI_Datatype type;
	COMM_INFO *cinfo = map->cinfo;

	if (nvec > 1) {
	    MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	    MPI_Type_commit(&type);
	}
	else {
	    type = PHG_MPI_FLOAT;
	}

	sbuff = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize) * sizeof(*sbuff));
	rbuff = sbuff + nvec * cinfo->ssize;

	/* copy data to send from data to sbuff */
	p = sbuff;
	for (i = 0; i < cinfo->ssize; i++) {
	    k = table[cinfo->sind[i]];
	    assert(k < vec->map->nlocal);
	    if (k == -1) {
		q = &zero;
		stride = 0;
	    }
	    else {
		q = vec->data + k;
		stride = vec->map->nlocal;
	    }
	    for (ii = 0; ii < nvec; ii++) {
		*(p++) = *q;
		q += stride;
	    }
	}

	phgAlltoallv(sbuff, cinfo->scnts, cinfo->sdsps, type,
		     rbuff, cinfo->rcnts, cinfo->rdsps, type, cinfo->comm);

	q = rbuff;
	offp_data = phgAlloc(nvec * cinfo->rsize * sizeof(*offp_data));
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data + cinfo->rind[i];
	    for (ii = 0; ii < nvec; ii++) {
		*p = *(q++);
		p += cinfo->rsize;
	    }
	}

	phgFree(sbuff);
	if (nvec > 1)
	    MPI_Type_free(&type);
    }
#endif	/* USE_MPI */

    for (id = 0; id < ndof; id++, dofs++) {
	for (iv = 0; iv < nvec; iv++)
	    phgDofClearCache(NULL, (*dofs)[iv], NULL, NULL, FALSE);
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    k = phgMapD2L(map, id, i);
	    if (k < 0 || (k = table[j = phgMapL2V(map, k)]) == -1) {
		data = &zero;
		stride = 0;
	    }
	    else if (k < vec->map->nlocal) {
		data = vec->data + k;
		stride = vec->map->nlocal;
	    }
	    else {
		assert(j >= map->nlocal);
		data = offp_data + j - map->nlocal;
		stride = map->localsize - map->nlocal;
	    }
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		u->data[i] = *data;
		data += stride;
	    }
	}
    }

    phgFree(offp_data);
    phgFree(table);
}

void
phgMapVecToDofArrays(MAP *map, VEC *vec, BOOLEAN remove_bdry, DOF **u, ...)
/* copies data from a vector to an array of DOFs */
{
    int ndof;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    phgMapVecToDofArraysN(map, vec, remove_bdry, ndof, dofs);
    phgFree(dofs);
}

VEC *
phgMapCreateVec(MAP *map, int nvec)
{
    VEC *v = phgCalloc(1, sizeof(*v));

    map->refcount++;

    v->magic = MAGIC_VEC;
    v->map = map;
    v->mat = NULL;

#if USE_MPI
    v->O2Gmap = NULL;
    v->ordering = NULL;
    v->localsize = 0;
    v->alloc = 0;
#endif	/* USE_MPI */

    v->assembled = TRUE;
    v->nvec = nvec;
    v->data = phgCalloc(nvec * map->nlocal, sizeof(*(v->data)));
    v->offp_data = phgCalloc(nvec * (map->localsize - map->nlocal),
						sizeof(*(v->data)));

    return v;
}

MAT *
phgMapCreateMat(MAP *rmap, MAP *cmap)
{
    MAT *Mp;

    FunctionEntry;

    Mp = phgCalloc(1, sizeof(*Mp));
    Mp->magic = MAGIC_MAT;
    Mp->rmap = rmap;
    Mp->cmap = cmap;
    Mp->type = PHG_UNPACKED;

    /* check for special matrix */
    if (Mp->rmap->nprocs == 1 && Mp->cmap->nprocs > 1) {
	phgPrintf("\n*\n* Warning: Matrices with serial rows and distributed "
		  "columns are inefficient.\n");
	phgPrintf("* Warning: It is better to use distributed rows and serial "
		  "columns instead.\n*\n");
    }

#if USE_MPI
    Mp->O2Gmap = NULL;
    Mp->ordering = NULL;
    Mp->cinfo = NULL;
#endif	/* USE_MPI */
    Mp->localsize = -1;	/* to make convert_row_index() initialize Mp->O2Gmap,
			 * etc., when needed. */

    Mp->rows = phgCalloc(rmap->localsize, sizeof(*Mp->rows));
    Mp->nnz_d = Mp->nnz_o = 0;

    Mp->diag = NULL;
    Mp->packed_data = NULL;
    Mp->packed_cols = NULL;

    Mp->mv_func = NULL;
    Mp->mv_data = NULL;

    Mp->assembled = FALSE;
    Mp->mode = PHG_UNDEFINED;

    rmap->refcount++;
    cmap->refcount++;

    Return Mp;

    return Mp;	/* to make MIPS cc happy */
}

MAT *
phgMapCreateMatrixFreeMat(MAP *rmap, MAP *cmap, MV_FUNC mv_func,
			  void *mv_data, ...)
{
    extern void _phg_matrix_free_setup(MAT *, MV_FUNC, int, void **);
    MAT *mat = phgMapCreateMat(rmap, cmap);
    int ndata;
    void **data;

    if (mat == NULL)
	return NULL;

    GetVariableArgs(ndata, data, mv_data, void *);
    _phg_matrix_free_setup(mat, mv_func, ndata, data);
    phgFree(data);

    return mat;
}
