/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: partition-utils.c,v 1.47 2009/05/01 05:19:48 zlb Exp $ */

#include "phg.h"
#include "partition-sfc.h"
#include "hamilton.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* load balance efficiency */
static const FLOAT SFC_LBE = 1.004;

/* ************************************************
 * Hilbert Order 
 * used by phgImport
 **************************************************/

/* ordering:
 * used for ordering Hilbert order and Morton order*/
static int
comp_key(const void *n1, const void *n2)
{
    SFC *m1, *m2;
    int flag = 0;

    m1 = (SFC *) n1;
    m2 = (SFC *) n2;
    if (m1->sfc > m2->sfc)
        flag = 1;
    else if (m1->sfc < m2->sfc)
        flag = -1;
    else 
         flag = 0;
    return flag;
}

/* generate Hilbert Order */
/* 0: success, or fail */
INT
phgHilbertOrder(GRID *g)
{
    SFC *x;
    INT i;
    FLOAT xmax, xmin;
    FLOAT ymax, ymin;
    FLOAT zmax, zmin;
    FLOAT etx, ety, etz;
    FLOAT x0, y0, z0;
    static const FLOAT lam[] = {0.25, 0.25, 0.25, 0.25};
    SIMPLEX *e;

    assert(g != NULL);
    if (g->nroot <= 0) {
        phgWarning("No Elements!\n");
        return 0;
    }
    if (g->nroot <= 2) {
        return 0;
    }

    x = (SFC *)phgAlloc(g->nroot * sizeof(SFC));
    if (x == NULL && g->nroot > 0) {
        phgError(1, "Can't allocate space: %s %d\n", __FILE__, __LINE__);
        return -1;
    }

    /* scan all elements */
    /* calculate barycentric coordinate */
    e = g->roots;
    x[0].index = e->index;
    phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
    x[0].co[0] = x0;          /* x */
    x[0].co[1] = y0;          /* y */
    x[0].co[2] = z0;          /* z */

    xmax = x[0].co[0];
    xmin = x[0].co[0];
    ymax = x[0].co[1];
    ymin = x[0].co[1];
    zmax = x[0].co[2];
    zmin = x[0].co[2];

    for (i = 1; i < g->nroot; i++) {
        e = g->roots + i;
        x[i].index = e->index;
        phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
        x[i].co[0] = x0;          /* x */
        x[i].co[1] = y0;          /* y */
        x[i].co[2] = z0;          /* z */
    } /* for */
    
    /* boundary box */
    xmin = g->bbox[0][0];
    ymin = g->bbox[0][1];
    zmin = g->bbox[0][2];
    xmax = g->bbox[1][0];
    ymax = g->bbox[1][1];
    zmax = g->bbox[1][2];

    /* scale bounding box  */
    /* such that x, y, z in [0, 1)^3 */
    etx = xmax - xmin;
    ety = ymax - ymin;
    etz = zmax - zmin;

    if (etx == 0.) etx = 1.;
    if (etx < ety) etx = ety;
    if (etx < etz) etx = etz;
    etx =  1. / etx;
    ety = etx;
    etz = etx;

    for (i = 0; i < g->nroot; i++) {
        x[i].co[0] = (x[i].co[0] - xmin) * etx;
        x[i].co[1] = (x[i].co[1] - ymin) * ety;
        x[i].co[2] = (x[i].co[2] - zmin) * etz;
    }
    /* generate Hilbert SFC */
    if (!phgSFCInvHilbert3D(x, g->nroot)) {
        phgWarning("Generating Hilbert SFC error: %s  %d\n", __FILE__, __LINE__);
        phgFree(x);
        return -1;
    }
    /* sort result */
    qsort(x, g->nroot, sizeof(SFC), comp_key);
    /* reorder the indices */
    {
        INT *aux;
        INT kk;

        aux = (INT *)phgAlloc(g->nroot * sizeof(INT));
        if (aux == NULL) {
            phgWarning("allocate space error: %s %d\n", __FILE__, __LINE__);
            phgFree(x);
            return -1;
        }

        for (i = 0; i < g->nroot; i++) {
            aux[x[i].index] = i;
        }

        for (i = 0; i < g->nroot; i++) {
            e = g->roots + i;
            kk = e->index;
            e->index = aux[kk];
        }
        phgFree(aux);
    }

    phgFree(x);
    return 0;
}

/* generate Morton Order */
/* 0: success, or fail */
INT
phgMortonOrder(GRID *g)
{
    SFC *x;
    INT i;
    FLOAT xmax, xmin;
    FLOAT ymax, ymin;
    FLOAT zmax, zmin;
    FLOAT etx, ety, etz;
    FLOAT x0, y0, z0;
    static const FLOAT lam[] = {0.25, 0.25, 0.25, 0.25};
    SIMPLEX *e;

    assert(g != NULL);
    if (g->nroot <= 0) {
        phgWarning("No Elements!\n");
        return 0;
    }
    if (g->nroot <= 2) {
        return 0;
    }

    x = (SFC *)phgAlloc(g->nroot * sizeof(SFC));
    if (x == NULL && g->nroot > 0) {
        phgError(1, "Can't allocate space: %s %d\n", __FILE__, __LINE__);
        return -1;
    }

    /* scan all elements */
    /* calculate barycentric coordinate */
    e = g->roots;
    x[0].index = e->index;
    phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
    x[0].co[0] = x0;          /* x */
    x[0].co[1] = y0;          /* y */
    x[0].co[2] = z0;          /* z */

    xmax = x[0].co[0];
    xmin = x[0].co[0];
    ymax = x[0].co[1];
    ymin = x[0].co[1];
    zmax = x[0].co[2];
    zmin = x[0].co[2];

    for (i = 1; i < g->nroot; i++) {
        e = g->roots + i;
        x[i].index = e->index;
        phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
        x[i].co[0] = x0;          /* x */
        x[i].co[1] = y0;          /* y */
        x[i].co[2] = z0;          /* z */
    } /* for */
    
    /* boundary box */
    xmin = g->bbox[0][0];
    ymin = g->bbox[0][1];
    zmin = g->bbox[0][2];
    xmax = g->bbox[1][0];
    ymax = g->bbox[1][1];
    zmax = g->bbox[1][2];

    /* scale bounding box  */
    /* such that x, y, z in [0, 1)^3 */
    etx = xmax - xmin;
    ety = ymax - ymin;
    etz = zmax - zmin;

    if (etx == 0.) etx = 1.;
    if (etx < ety) etx = ety;
    if (etx < etz) etx = etz;
    etx =  1. / etx;
    ety = etx;
    etz = etx;

    for (i = 0; i < g->nroot; i++) {
        x[i].co[0] = (x[i].co[0] - xmin) * etx;
        x[i].co[1] = (x[i].co[1] - ymin) * ety;
        x[i].co[2] = (x[i].co[2] - zmin) * etz;
    }
    /* generate Morton SFC */
    if (!phgSFCInvMorton3D(x, g->nroot)) {
        phgWarning("Generating Morton SFC error: %s  %d\n", __FILE__, __LINE__);
        phgFree(x);
        return -1;
    }
    /* sort result */
    qsort(x, g->nroot, sizeof(SFC), comp_key);
    /* reorder the indices */
    {
        INT *aux;
        INT kk;

        aux = (INT *)phgAlloc(g->nroot * sizeof(INT));
        if (aux == NULL) {
            phgWarning("allocate space error: %s %d\n", __FILE__, __LINE__);
            phgFree(x);
            return -1;
        }

        for (i = 0; i < g->nroot; i++) {
            aux[x[i].index] = i;
        }

        for (i = 0; i < g->nroot; i++) {
            e = g->roots + i;
            kk = e->index;
            e->index = aux[kk];
        }
        phgFree(aux);
    }

    phgFree(x);
    return 0;
}

/* construct initial grid order 
 * default: hilbert order */
static const char *orders[] = {"hilbert", "morton", "hamilton", "none", NULL};
static int GridOrder = 0;

INT
phgGridInitOrder(GRID *g, BOOLEAN dist)
{
    INT ret = 1;
    FLOAT t1[3], t2[3];
    static BOOLEAN initialized = FALSE;

    /* register parameters */
    if (!initialized) {
        initialized = TRUE;
        phgOptionsRegisterTitle("RTK options:", NULL, "partition");
        phgOptionsRegisterKeyword("-rtk_init_order", "Initial ordering",
                orders, &GridOrder);
        return TRUE;
    }

    assert(g != NULL);
    if (dist) {  /* parallel version */
        phgError(1, "Not implemented yet!\n");
    }
    else {       /* sequential version */
        if (phgVerbosity > 0) {
            phgPrintf("Initial ordering for RTK: ");
            phgGetTime(t1);
        }
        
        /* Hilbert order */
        if (GridOrder == 0) {
            if (phgVerbosity > 0) {
                phgPrintf("Hilbert space-filling curve.\n");
            }
            ret = phgHilbertOrder(g);
        } /* Morton order */
        else if (GridOrder == 1) {
            if (phgVerbosity > 0) {
                phgPrintf("Morton space-filling curve.\n");
            }
            ret = phgMortonOrder(g);
        }    /* Hamilton Path */
        else if (GridOrder == 2) {
            if (phgVerbosity > 0) {
                phgPrintf("\tHamilton path.\n");
            }
            ret = phgHamiltonPath(g);
        }
	else {
            if (phgVerbosity > 0) {
                phgPrintf("none.\n");
            }
	    ret = 0;
	}
        
        /* success */
        if (ret == 0) {
            if (phgVerbosity > 0) {
                phgGetTime(t2);
                phgPrintf("Finished computing initial ordering of elements: ");
                phgPrintf("%fs\n", t2[2] - t1[2]);
                phgPrintf("Total roots: %d\n", g->nroot);
            }
        }
        else { /* fail */
            if (phgVerbosity > 0) {
                phgPrintf("Error in computing initial ordering.");
            }
        }
    }
    return ret;
}

#if USE_MPI
/***************************************************
 *             1D partitioner 
 **************************************************/
typedef struct {
    FLOAT low;
    FLOAT high;
    INT   pn;
}PTNS;

/* tolerance for load balance efficiency */
static FLOAT LB_TOL;

/* return the partition id that key belongs to */
static INT bsearch_ptn(PTNS *ptn, INT p, FLOAT key);

#define CUTS 16

/* cut [0, 1] to p parts, such that lbe <= LB_TOL 
 * dts: store key \in [0, 1]
 * lx: length of dts 
 * p: # of partitions */
FLOAT
phgPartition1D(DOTS *dts, INT lx, INT p, MPI_Comm comm)
{
    INT npct = (p - 1) * CUTS + 1;  /* total  cuts - 1 */
    PTNS  ptn[npct];

    FLOAT lbe;                      /* load balance efficiency */
    FLOAT DotsSum;                  /* total sum */
    FLOAT psum[p];                  /* partial sum, psum[i] = sf * i * sum */
    FLOAT sum[npct];
    FLOAT tsum[npct];
    FLOAT bbox[p][2];               /* bounding box, 0: low, 1: high*/

    INT i,j;
    INT ctn;                        /* if we start next loop */
    INT rank;                       /* pid */
    INT flag;                       /* iterative times */
    INT map[npct];                  /* partition result */
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
        initialized = TRUE;
        LB_TOL = SFC_LBE;
        phgOptionsRegisterTitle("Options for 1D partitioner:", NULL, 
                "partition");
        phgOptionsRegisterFloat("-lbe_tol", "Tolerance of load balance "
		"efficiency (>= 1.)", &LB_TOL);

        return 0.;
    }

    if (LB_TOL < 1.) {
	phgWarning("LB_TOL too small, changed to %lg\n", (double)SFC_LBE);
        LB_TOL = SFC_LBE;
    }
    else if (LB_TOL > 1.3) {
	phgWarning("LB_TOL too large (poor load balance)\n");
    }

    assert(lx >= 0 && p >= 1);
    if (lx > 0) {
        assert(dts != NULL);
    }

    /* deal with special cases */
    if (p == 1) {  /* only one partition */
        for (i = 0; i < lx; i++) {
            dts[i].pn = 0;
        }
        return 1.;
    }
    
    /* no dot, just return */
    rank = lx;
    MPI_Allreduce(&rank, &ctn, 1, PHG_MPI_INT, MPI_SUM, comm);
    if (ctn == 0) {
        return 1.;
    }

    /* total sum */
    lbe = 0.;
    for (i = 0; i < lx; i++) {
        lbe += dts[i].w;
    }
    MPI_Allreduce(&lbe, &DotsSum, 1, PHG_MPI_FLOAT, MPI_SUM, comm);

    /* special case */
    if (DotsSum == 0.) {
        for (i = 0; i < lx; i++)
            dts[i].pn = 0;
        return 1.;
    }
    else if (DotsSum < 0.) {
        phgError(1, "negative weights %s  %d\n", __FILE__, __LINE__);
    }

    /* initialize partial sum, ideal condition  */
    /* initialize bounding boxs */
    lbe = 1.0 / p;
    ctn = p - 1;
    for (i = 0; i < ctn; i++) {
        psum[i] = DotsSum * (i + 1.) * lbe;
        bbox[i][0] = 0.;
        bbox[i][1] = 1.;
    }
    psum[ctn] = DotsSum;
    bbox[ctn][0] = 0.;
    bbox[ctn][1] = 1.;

    /* initial guess partitions */
    lbe = 1. / npct;
    for (i = 0; i < npct; i++) {
        ptn[i].low = i * lbe;
        ptn[i].high = (i + 1) * lbe;
        ptn[i].pn = i;
    }

    MPI_Comm_rank(comm, &rank);

    /* partitioning 
     * phgRank 0 generates results then broadcast */
    flag = 0;
    ctn = 1;
    while (flag < MAXLOOPS && ctn) {
        FLOAT mxm;
        FLOAT tmp[p];
        INT ct[npct];  /* record if current interval is overflow interval */
                       /* 0: no; > 0, yes and number of overflow intervals */
        INT p1 = p - 1;

        /* sum of current partitions */
        for (i = 0; i < npct; i++) tsum[i] = 0.;
        for (i = 0; i < lx; i++) {
            ctn = bsearch_ptn(ptn, npct, dts[i].key);
            dts[i].pn = ctn;
            tsum[ctn] += dts[i].w;
        }

        /* calculate total weights of all intervals */
        MPI_Reduce(tsum, sum, npct, PHG_MPI_FLOAT, MPI_SUM, 0, comm);

        /* rank == 0 generate results */
        if (rank == 0) {
            /* prefix sum */
            tsum[0] = sum[0];
            ct[0] = 0;
            for (i = 1; i < npct; i++) {
                tsum[i] = sum[i];
                sum[i] += sum[i - 1];
                ct[i] = 0;
            }

            /* update bounding box */
            j = 0;
            for (i = 0; i < p1; i++) {
                /* look for new bounding box position */
                while (sum[j] <= psum[i]) {
                    j++;
                }
                /* update bounding box */
                if (ptn[j].high < bbox[i][1]) bbox[i][1] = ptn[j].high;
                if (ptn[j].low  > bbox[i][0]) bbox[i][0] = ptn[j].low;
                ct[j] += 1;

                tmp[i] = 0.;
            }
            tmp[p - 1] = 0.;

            /* calculate max sum of current partition */
            /* map map[npct] to p parts */
            j = 0;
            for (i = 0; i < npct && j < p; i++) {
                if (sum[i] < psum[j]) {
                    map[i] = j;
                }
                else {
                    map[i] = j;
                    j++;
                }
            }
            /* special case that j >= p */
            for (j = i; j < npct; j++) {
                map[j] = p1;
            }

            for (i = 0; i < npct; i++) {
                j = map[i];
                tmp[j] += tsum[i];
            }
            mxm = tmp[0];
            for (i = 1; i < p; i++) {
                if (mxm < tmp[i]) mxm = tmp[i];
            }

            /* calculate load balance efficiency 
             * quit if satisfies conditon, or continue */
            lbe = mxm * p / DotsSum;
            if (lbe <= LB_TOL) {
                ctn = 0;
            }
            else {
                ctn = 1;
            }
        }

        /* synchronize status */
        /* 1: start next iteration, 0: stop */
        MPI_Bcast(&ctn, 1, PHG_MPI_INT, 0, comm);
       
        if (ctn) {
            ctn = npct - 1;

            /* generate new partition */
            /* cut p - 1 bounding boxes to npct parts */
            if (rank == 0) {
                FLOAT len;
                INT k, j1, j2;
                INT ctk;
                
                i = 0;
                j = 0;
                k = 0;
                while (i < p1) {
                    while (ct[j] < 1) {
                        j++;
                    }

                    /* new partition 
                     * we use array sum[ctn] for temp variable */
                    j2 = ct[j] * CUTS;
                    if (j < ctn) {
                        len = 1. / (j2 + 1.);

                        ctk = i + ct[j] - 1;
                        for (j1 = 1; j1 <= j2; j1++) {
                            sum[k] = (bbox[i][0] * (j2 + 1 - j1) + bbox[ctk][1] * j1) * len;
                            k++;
                        }
                    }
                    else {
                        len = 1. / (j2 + 2.);
                        ctk = i + ct[j] - 1;
                        for (j1 = 1; j1 <= j2; j1++) {
                            sum[k] = (bbox[i][0] * (j2 + 2 - j1) + bbox[ctk][1] * j1) * len;
                            k++;
                        }

                    }

                    i += ct[j];
                    j++;
                }
#if 0
                /* check */
                for (i = 1; i < ctn; i++) {
                    assert(sum[i - 1] <= sum[i]);
                }
#endif
            }

            /* broadcast new partitions */
            /* initialize new partitions */
            MPI_Bcast(sum, ctn, PHG_MPI_FLOAT, 0, comm);

            ptn[0].low = 0.;
            ptn[0].high = sum[0];
            for (i = 1; i < ctn; i++) {
                ptn[i].high = sum[i];
                ptn[i].low = sum[i - 1];
            }
            ptn[ctn].high = 1.;
            ptn[ctn].low = sum[ctn - 1];
            
            ctn = 1;
        }
        else {
            break;
        }

        flag++;
    }  /* while */

    /* synchronize results */
    MPI_Bcast(map, npct, PHG_MPI_INT, 0, comm);
    MPI_Bcast(&lbe, 1, PHG_MPI_FLOAT, 0, comm);

    for (i = 0; i < lx; i++) {
        j = dts[i].pn;
        dts[i].pn = map[j];
    }

    if (phgVerbosity > 0) {
        phgPrintf("1D partitioner iterations: %d\n", flag);
        phgPrintf("Load imbalance factor: %.8f\n", lbe);
    }

    return lbe;
}

static INT
bsearch_ptn(PTNS *ptn, INT p, FLOAT key)
{
    INT lo, hi;
    INT md;
    INT ret;
    
    assert(p >= 1);
    
    lo = 0;
    hi = p - 1;
    md = (lo + hi) >> 1;
    while (1) {
        if (key >= ptn[md].high) ret = 1;
        else if (key - ptn[md].low < 0.) ret = -1;
        else ret = 0;

        if (ret == 1) {
            lo = md;
            md = (lo + hi) >> 1;
            if (md == lo) md = hi;
        }
        else if (ret == -1) {
            hi = md;
            md = (lo + hi) >> 1;
        }
        else {
            break;
        }
    }
    return md;
}

/* phgPartitionRemap, remap the partitions such that the
 * distributed data will be small
 * nprocs: partition number
 * pr: result. Map i to pr[i]
 * if TRUE, partitions have been remapped
 * if FALSE, partitions keep the same
 *
 * The algorithm comes from PLUM: 
 * parallel load balancing for adaptive unstructured meshed 
 * Leonid Oliker, Rupak Biswas 
 *
 * heuristic algorithm */

/* similiar matrix */
typedef struct {
    INT s;      /* number of elements */
    int pid;    /* current pid */
    int ptn;    /* partition id */
} SM;

/* descending order used by qsort */
static int
comp_sml(const void *m1, const void *m2)
{
    SM *n1, *n2;
    int flag = 0;
    
    n1 = (SM *)m1;
    n2 = (SM *)m2;

    if (n1->s > n2->s)
        flag = -1;
    else if (n1->s == n2->s)
        flag = 0;
    else if (n1->s < n2->s)
        flag = 1;
    return flag;
}

BOOLEAN
phgPartitionRemap(GRID *g, INT nprocs, MPI_Comm comm)
{
    SM *local, *sml;
    SIMPLEX *e;
    MPI_Datatype type;
    MPI_Datatype types[3];
    MPI_Aint indices[3];
    INT i, j;
    int *result;
    int rank, size;
    int blocklens[3];
    BOOLEAN ret = FALSE;

    i = g->nleaf;
    MPI_Allreduce(&i, &j, 1, PHG_MPI_INT, MPI_SUM, comm);
    if ((j == 0) || (nprocs <= 1)) {
        return FALSE;
    }

    local = (SM *) phgAlloc(nprocs * sizeof(SM));
    sml = (SM *) phgAlloc(nprocs * nprocs * sizeof(SM));
    result = (int *) phgAlloc(nprocs * sizeof(int));

    if ((local == NULL) || (sml == NULL) || (result == NULL)) {
        phgError(1, "can't locate spaces %s  %d\n", __FILE__, __LINE__);
    }

    /* create a new datatype for exchange information */
    /* used by structure tree */
    blocklens[0] = blocklens[1] = blocklens[2] = 1;
    types[0] = PHG_MPI_INT;
    types[1] = types[2] = MPI_INT;
    MPI_Address(&local[0].s, &indices[0]);
    MPI_Address(&local[0].pid, &indices[1]);
    MPI_Address(&local[0].ptn, &indices[2]);
    indices[2] -= indices[0];
    indices[1] -= indices[0];
    indices[0] = 0;
    MPI_Type_struct(3, blocklens, indices, types, &type);
    MPI_Type_commit(&type);

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    /* gather information to creat similiar matrix */
    for (i = 0; i < nprocs; i++) {
        local[i].s = 0;
        local[i].pid = rank;
        local[i].ptn = i;
    }
    ForAllElements(g, e) {
        local[e->mark].s += 1;
    }
    MPI_Gather(local, nprocs, type, sml, nprocs, type, 0, comm);
    MPI_Type_free(&type);

    /* map algorithm by PLUM */
    if (rank == 0){
        int *part_map;
        int *proc_unmap;
        int *t1;
        int count;
        SM *p;

        part_map = (int *) phgAlloc(nprocs * sizeof(int));
        proc_unmap = (int *) phgAlloc(nprocs * sizeof(int));
        t1 = (int *) phgAlloc(nprocs * sizeof(int));

        if ((part_map == NULL) || (proc_unmap == NULL) || (t1 == NULL)) {
            phgError(1, "can't locate spaces %s  %d\n", __FILE__, __LINE__);
        }

        /* inlarge similiar matrix */
        /* such that sml is square */
        if (size < nprocs) {
            for (i = size; i < nprocs; i++) {
                for (j = 0; j < nprocs; j++) {
                    count = i * nprocs + j;
                    sml[count].pid = i;
                    sml[count].ptn = j;
                    sml[count].s = 0;
                }
            }
        }

        /* initialize */
        for (i = 0; i < nprocs; i++) {
            proc_unmap[i] = 1;
            part_map[i] = 0;
        }
        /* order the similiar matrix, descending order */
        qsort(sml, size * nprocs, sizeof(SM), comp_sml);

        /* map algorithm */
        count = 0;
        p = sml;
        while (count < nprocs) {
            while ((proc_unmap[p->pid] == 0 || part_map[p->ptn] == 1)) {
                p++;
            }
            proc_unmap[p->pid] = 0;
            part_map[p->ptn] = 1;
            result[p->ptn] = p->pid;
            count++;
        } /* map algorithm */

        /* verification */
        /* TRUE: valid result   */
        /* FALSE: invalid result. unmap. */
        for (i = 0; i < nprocs; i++)
            t1[i] = 0;
        for (i = 0; i < nprocs; i++)
            t1[result[i]] += 1;
        count = 0;
        for (i = 0; i < nprocs; i++) {
            if (t1[i] != 1)
                count++;
        }

        if (count == 0) {
            ret = TRUE;
        }
        else {
            ret = FALSE;
            for (i = 0; i < nprocs; i++)
                result[i] = i;
        }

        phgFree(part_map);
        phgFree(proc_unmap);
        phgFree(t1);
    }

    /* broadcast result and map*/
    MPI_Bcast(&ret, 1, MPI_INT, 0, comm);
    if (ret == TRUE) {
        MPI_Bcast(result, nprocs, MPI_INT, 0, comm);
        ForAllElements(g, e) {
            i = e->mark;
            e->mark = result[i];
        }
    }

    phgFree(sml);
    phgFree(local);
    phgFree(result);

    return ret;
}

/* print grid quality information */
/* this function must be called just after phgBalance */
/* or e->mark isn't changed.   */
void
phgPartitionQuality(GRID *g)
{
    INT i, k;
    INT cr, ci, cb;
    INT tcr, tci, tcb;
    INT en;
    FLOAT si, tsi;
    SIMPLEX *e;
    INT *pid;

    if (g->nroot <= 0)
        return;
    if (phgRank >= g->nprocs)
        return;

    en = ci = cb = cr = 0;
    ForAllElements(g, e) {
        en++;
        for (i = 0; i < NFace; i++) {
            if (e->bound_type[i] & REMOTE) {
                cr++;
                continue;
            }
            if (e->bound_type[i] & INTERIOR) {
                ci++;
                continue;
            }
            cb++;
        }
    }

    /* load imbalance efficiency */
    MPI_Allreduce(&en, &k, 1, PHG_MPI_INT, MPI_MAX, g->comm);
    si = (FLOAT)k * g->nprocs / (FLOAT)g->nelem_global;
    phgPrintf("g->lif:    %f\n", si);

    /* total cut edges */
    MPI_Allreduce(&cr, &tcr, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    tcr >>= 1;
    phgPrintf("Cut edges: %d\n", tcr);

    /* surface indics */
    ci >>= 1;
    k = cb + ci + cr;
    /* local surface index */
    if (k > 0) {
        si = (FLOAT) cr / (FLOAT) k;
    }
    else {
        si = 0;
    }
    MPI_Allreduce(&si, &tsi, 1, PHG_MPI_FLOAT, MPI_MAX, g->comm);
    phgPrintf("Maximum local surface index: %f\n", tsi);

    /* average surface index */
    MPI_Allreduce(&si, &tsi, 1, PHG_MPI_FLOAT, MPI_SUM, g->comm);
    si = tsi / (FLOAT)g->nprocs;
    phgPrintf("Average surface index:       %f\n", si);

    /* global surface index */
    MPI_Allreduce(&cb, &tcb, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    MPI_Allreduce(&ci, &tci, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    k = tci + tcr + tcb;

    si = (FLOAT) tcr / (FLOAT)k;
    phgPrintf("Global surface index:        %f\n", si);

    /* compute interprocess connectivity */
    pid = phgAlloc((g->nprocs + 1) * sizeof(*pid));
    for (i = 0; i < g->nprocs; i++) {
        if (i != phgRank && g->neighbours.counts[i] != 0) {
            pid[i] = 1;
        }
        else {
            pid[i] = 0;
        }
    }
    k = 0;
    for (i = 0; i < g->nprocs; i++)
        k += pid[i];
    phgFree(pid);
    MPI_Allreduce(&k, &i, 1, PHG_MPI_INT, MPI_MAX, g->comm);
    phgPrintf("Max interprocess connectivity: %d\n", i);
}
#endif
